#include <stdio.h>     /* For perror() */
#include <unistd.h>    /* For write()  */
#include <time.h>

#include "stdin.hh"
#include "datapacket.hh"
#include "connectioncache.hh"
#include "mastermsg.hh"

StdinType::StdinType() : cache(new ConnectionCache(this)), ExpectLen(0)
{
}

StdinType::~StdinType()
{
    if(cache)delete cache;
}

void StdinType::Die()
{
    if(cache) { delete cache; cache = NULL; }
}

bool StdinType::IsDead() const
{
    return cache == NULL;
}

void StdinType::AddOpeningMsg(unsigned ConnID, SockProtocol proto, int port)
{
    MasterMsg m(MasterMsg::MsgTypeOpen);
    
    m.SetParam(MasterMsg::ParamTypeConnID, Int2Data(ConnID));
    m.SetParam(MasterMsg::ParamTypePort,   Int2Data(port));
    m.SetParam(MasterMsg::ParamTypeProto,  Proto2Data(proto));
    
    AddMsg(m);
}

void StdinType::AddClosingMsg(unsigned ConnID)
{
    MasterMsg m(MasterMsg::MsgTypeClose);

    m.SetParam(MasterMsg::ParamTypeConnID, Int2Data(ConnID));

    AddMsg(m);
}

void StdinType::AddDataMsg(unsigned ConnID, const vector<Datatype> &data)
{
    MasterMsg m(MasterMsg::MsgTypeData);
    
    m.SetParam(MasterMsg::ParamTypeConnID, Int2Data(ConnID));
    m.SetParam(MasterMsg::ParamTypeData,   DatatableType(data.begin(), data.size()));

    AddMsg(m);
}

void StdinType::AddListenMsg(SockProtocol proto, int port)
{
    MasterMsg m(MasterMsg::MsgTypeListen);
    
    m.SetParam(MasterMsg::ParamTypePort,  Int2Data(port));
    m.SetParam(MasterMsg::ParamTypeProto, Proto2Data(proto));
    
    AddMsg(m);
}

void StdinType::AddMsg(const MasterMsg &m)
{
    DatatableType data = m.MakeString();
    DatatableType size = Int2Data(data.size());
    
    SendQueue.Feed(size.begin(), size.size());
    SendQueue.Feed(data.begin(), data.size());
}

void StdinType::Speak(int fd)
{
    const Datatype *buf = SendQueue.GetBegin();
    unsigned        len = SendQueue.GetSize();
    
    int status = write(fd, buf, len);
    if(status < 0)
    {
        /* FIXME: What should we do instead? */
        perror("write");
        return;
    }
    
    if(status > 0)
        SendQueue.Eat(status);
}

DataPacket StdinType::Read(int fd)
{
    DataPacket pak;
    
    if(!ExpectLen)
    {
        Datatype Arr[4];
        int status = read(fd, Arr, sizeof Arr);
        if(status < 0)
        {
            /* FIXME: What should we do instead? */
            perror("read");
            return pak;
        }
        if(status < 4)
        {
            /* FIXME: This *should* work too. */
            return pak;
        }
        ExpectLen = Data2Int(DatatableType(Arr, sizeof Arr));
        
        if(ExpectLen > 200000)
            ExpectLen = 0;
        if(ExpectLen > 0)
            ExpectBuf = new Datatype[ExpectLen];
        ExpectGot = 0;
    }
    else
    {
        unsigned Left = ExpectLen;
        if(Left > ExpectGot)Left -= ExpectGot;
        
        fprintf(stderr, "Full size %u, got %u, trying %u\n", ExpectLen, ExpectGot, Left);

        int status = read(fd, ExpectBuf+ExpectGot, Left);
        if(status < 0)
        {
            /* FIXME: What should we do instead? */
            perror("read");
            return pak;
        }
        ExpectGot += status;
        if(ExpectGot == ExpectLen)
        {
            MasterMsg m(DatatableType(ExpectBuf, ExpectLen));
            
            /* Master said its words. Let's heed them. */
            switch(m.GetType())
            {
                case MasterMsg::MsgTypeListen:
                {
                	/* Sender: Master */
                    pak.Type          = DataPacket::PacketTypeListen;
                    
                    pak.Peer.Protocol = Char2Proto(m.GetCharParam(MasterMsg::ParamTypeProto));
                    pak.Peer.DestPort = m.GetIntParam(MasterMsg::ParamTypePort);
                    break;
                }
                case MasterMsg::MsgTypeClose:
                {
                	/* Sender: Master or beacon */
                	
                    unsigned ConnID = m.GetIntParam(MasterMsg::ParamTypeConnID);
                    
                    if(cache->Has(ConnID))
                    {
	                    pak.Type   = DataPacket::PacketTypeClose;
    	                pak.Peer   = cache->GetConnection(ConnID);
    	            }
                    break;
                }
                case MasterMsg::MsgTypeData:
                {
                	/* Sender: Master or beacon */
                	
                    unsigned ConnID = m.GetIntParam(MasterMsg::ParamTypeConnID);
                    
                    if(cache->Has(ConnID))
                    {
                        pak.Type   = DataPacket::PacketTypeData;
                        pak.Peer   = cache->GetConnection(ConnID);
                        
                        cache->TouchConnection(ConnID);
                        
                        DatatableType data = m.GetDataParam(MasterMsg::ParamTypeData);
                        pak.MakeData(data.begin(), data.size());
                    }
                    
                    break;
                }
                case MasterMsg::MsgTypeOpen:
                {
                	/* Sender: Beacon */
                	
                	unsigned ConnID = m.GetIntParam(MasterMsg::ParamTypeConnID);

                    pak.Type          = DataPacket::PacketTypeOpen;
                    pak.Peer.Protocol = Char2Proto(m.GetCharParam(MasterMsg::ParamTypeProto));
                    pak.Peer.DestPort = m.GetIntParam(MasterMsg::ParamTypePort);
                    pak.ConnID        = ConnID;
                	
                	break;
                }
                default:
                    /* FIXME: Somebody speaks nonsense! What to do now? */
                    break;
            }
            
            /* Delete this message, prepare for next. */
            delete[] ExpectBuf;
            ExpectLen = 0;
        }
    }
    return pak;
}


void StdinType::SendPacket(const struct DataPacket &pak)
{
    switch(pak.Type)
    {
        case DataPacket::PacketTypeNothing:
        {
            /* Do not bother the master with nothingness. */
            /* FIXME: Is this ok? Could there be a reason for idle messages? */
            break;
        }
        case DataPacket::PacketTypeOpen:
        {
            unsigned ConnID = cache->GetConnectionID(pak.Peer);
            
            /* Do not send message. GetConnectionID() does it. */
            /* ConnID therefore is unused! */
            ConnID=ConnID;
            break;
        }
        case DataPacket::PacketTypeData:
        {
            unsigned ConnID = cache->GetConnectionID(pak.Peer);
            
            /* Send the data message. */
            AddDataMsg(ConnID, pak.Data);
            
            break;
        }
        case DataPacket::PacketTypeClose:
        {
            unsigned ConnID = cache->GetConnectionID(pak.Peer);
            
            cache->DeleteConnectionID(ConnID);

            /* Do not send message. DeleteConnectionID() does it. */
            break;
        }
        case DataPacket::PacketTypeListen:
        {
            AddListenMsg(pak.Peer.Protocol, pak.Peer.DestPort);
            break;
        }
    }
}

bool StdinType::WantSpeak() const
{
    return SendQueue.GetSize();
}

void StdinType::AssignConnID(unsigned id, const Connection &conn)
{
	cache->Assign(id, conn);
}
